/*****************************************
 * AlKryptoKer
 * par Alain VIZZINI (2002-04-28)
 *
 * Classes :
 *       AlKryptoKer
 *       AKKPassword
 *       InvalidDESPassword
 *       BadPassword
 *****************************************/

using System;
using System.IO;
using System.Security;
using System.Security.Cryptography;
using System.Text; 

using K = System.Console;


///<summary>Kernel de AlKrypto</summary>
///<remark>Ce programme a t largement inspir de 2 sources : KB de MS et un article de MasterC#</remark>
public class AlKryptoKer {
  private static int Iter_Limit = 1048576; //1Mo
  
  ///<summary>La limite en taille du flux au-dessus de laquelle on utilise le mode iter plutot que une passe</summary>
  public static int ITER_LIMIT{
    get{
      return Iter_Limit;
    }
    set{
      if (value >= 0)
	Iter_Limit = value;
    }
  }

  ///<summary>Constructeur</summary>
  public AlKryptoKer(){}

  //-----------------------------------------------------------------------------------------------------------------
  //--                                                 Crypt                                                       --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Decrypt un fichier encod en une passe</summary>
  ///<remark>A utiliser pour les fichiers de taille relative (utiliser la version itrative pour les gros)</remark>
  public static void Crypt (AKKPassword s, String fichierIN, String fichierOUT){
    FileStream fin = new FileStream(fichierIN, 
				    FileMode.Open, 
				    FileAccess.Read);

    FileStream fout = new FileStream(fichierOUT, 
				     FileMode.Create, 
				     FileAccess.Write);
    try {
      if (fin.Length >= Iter_Limit)
	//gros fichier --> iteratif
	CryptFluxIter (s, fin, fout);
      else
	//petit fichier --> un bloc
	CryptFlux (s, fin, fout);
    }
    finally{
      fin.Close();
      fout.Close();
    }
  }


  //-----------------------------------------------------------------------------------------------------------------
  //--                                                Decrypt                                                      --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Decrypt un fichier encod en une passe</summary>
  ///<remark>A utiliser pour les fichiers de taille relative (utiliser la version itrative pour les gros)</remark>
  public static void DeCrypt (AKKPassword s, String fichierIN, String fichierOUT){
    FileStream fin = new FileStream(fichierIN, 
				    FileMode.Open, 
				    FileAccess.Read);

    FileStream fout = new FileStream(fichierOUT, 
				     FileMode.Create, 
				     FileAccess.Write);

    try{
      if (fin.Length >= Iter_Limit)
	//gros fichier --> iteratif
	DecryptFluxIter (s, fin, fout);
      else
	//petit fichier --> un bloc
	DecryptFlux (s, fin, fout);
    }
    catch (CryptographicException){
      throw new BadPassword();
    }
    finally {
      fin.Close();
      fout.Close();
    }
  }

  //-----------------------------------------------------------------------------------------------------------------
  //--                                                 Crypt                                                       --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Decrypt un fichier encod en une passe</summary>
  ///<remark>les flux doivent tre ouverts, on doit pouvoir crire dans streamOUT et les flux sont toujours ouverts aprs appel</remark>
  public static void Crypt (AKKPassword s, Stream streamIN, Stream streamOUT){
    if (streamIN.Length >= Iter_Limit)
      //gros fichier --> iteratif
      CryptFluxIter (s, streamIN, streamOUT);
    else
      //petit fichier --> un bloc
      CryptFlux (s, streamIN, streamOUT);
  
  }


  //-----------------------------------------------------------------------------------------------------------------
  //--                                                Decrypt                                                      --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Decrypt un fichier encod en une passe</summary>
  ///<remark>les flux doivent tre ouverts, on doit pouvoir crire dans streamOUT et les flux sont toujours ouverts aprs appel</remark>
  public static void DeCrypt (AKKPassword s, Stream streamIN, Stream streamOUT){
    try{
      if (streamIN.Length >= Iter_Limit)
	//gros fichier --> iteratif
	DecryptFluxIter (s, streamIN, streamOUT);
      else
	//petit fichier --> un bloc
	DecryptFlux (s, streamIN, streamOUT);
    }
    catch (CryptographicException){
      throw new BadPassword();
    }
  }
  


  //-----------------------------------------------------------------------------------------------------------------
  //--                                                 M A I N                                                     --
  //-----------------------------------------------------------------------------------------------------------------
  public static void Main(String [] a){
     K.WriteLine("*** AlKrypterKer(nel) ***");
     /*
       //exemples ..
       AKKPassword p = new AKKPassword (s);
       
       AlKryptoKer.Crypt (p, @"c:\dir_al\telephone.txt", "telk.txt");
       
       AlKryptoKer.DeCrypt (new AKKPassword ("iuytreza"), "telk.txt", "tel.txt");     
       
       AlKryptoKer.Crypt (p, "n.jpg", "n2.jpg");
       AlKryptoKer.DeCrypt (p, "n2.jpg", "n3.jpg");     
     */
  }



  //-----------------------------------------------------------------------------------------------------------------
  //--                                               CryptFlux                                                     --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Crypt un flux encod en une passe</summary>
  ///<remark>A utiliser pour les flux de petite taille. Les flux sont NON ferms  la fin de l'exec</remark>
  private static void CryptFlux (AKKPassword s, Stream sIN, Stream sOUT){
    DESCryptoServiceProvider DES = new DESCryptoServiceProvider(); 

    //DES utilise une cl (pwd) & un I(nitialization) V(ector) : mme contraire = 8 char
    DES.Key = s.PwdBytes();
    DES.IV = s.PwdBytes();

    //on cre un encrypteur (k) et un flux crypt
    ICryptoTransform k = DES.CreateEncryptor();
    CryptoStream cryptostream = new CryptoStream(sOUT, 
						 k, 
						 CryptoStreamMode.Write); 
    
    
    //un buffer de byte de la taille du fichier input !
    byte[] bytearrayinput = new byte[sIN.Length];
    int lu = sIN.Read(bytearrayinput, 0, bytearrayinput.Length);
    cryptostream.Write(bytearrayinput, 0, lu); 

    cryptostream.Close();
  }

  //-----------------------------------------------------------------------------------------------------------------
  //--                                             DecryptFlux                                                     --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Decrypt un flux encod en une passe</summary>
  ///<remark>A utiliser pour les flux de petite taille. Les flux sont NON ferms  la fin de l'exec</remark>
  private static void DecryptFlux (AKKPassword s, Stream sIN, Stream sOUT){
    DESCryptoServiceProvider DES = new DESCryptoServiceProvider(); 

    //DES utilise une cl (pwd) & un I(nitialization) V(ector) : mme contraire = 8 char
    DES.Key = s.PwdBytes();
    DES.IV = s.PwdBytes();

    ICryptoTransform dk = DES.CreateDecryptor();
    CryptoStream cryptostream = new CryptoStream(sIN,
						 dk, 
						 CryptoStreamMode.Read); 
    
    //un buffer de byte de la taille du fichier input !
    byte[] bytearrayinput = new byte[sIN.Length];
    int lu = cryptostream.Read(bytearrayinput, 0, bytearrayinput.Length);
    sOUT.Write(bytearrayinput, 0, lu); 

    cryptostream.Close();
  }

  //-----------------------------------------------------------------------------------------------------------------
  //--                                           CryptFluxIter                                                     --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Crypt un flux encod en plusieurs passe</summary>
  ///<remark>A utiliser pour les flux de grande taille. Les flux sont NON ferms  la fin de l'exec</remark>
  private static void CryptFluxIter (AKKPassword s, Stream sIN, Stream sOUT){
    DESCryptoServiceProvider DES = new DESCryptoServiceProvider(); 

    //DES utilise une cl (pwd) & un I(nitialization) V(ector) : mme contraire = 8 char
    DES.Key = s.PwdBytes();
    DES.IV = s.PwdBytes();

    //on cre un encrypteur (k) et un flux crypt
    ICryptoTransform k = DES.CreateEncryptor();
    CryptoStream cryptostream = new CryptoStream(sOUT, 
						 k, 
						 CryptoStreamMode.Write); 
    
    //** la diffrence est ici : on fait des itrations ... **
    
    //le buffer sera de taille multiple de la taille de la cl (8 octets) -> 16384
    byte[] bytearrayinput = new byte[16384];
    int effectue = 0;
    int lu;
    //FichierCrypt.Taille >= FichierNONCrypt.Taille
    while (effectue < sIN.Length){
      lu = sIN.Read(bytearrayinput, 0, bytearrayinput.Length);
      cryptostream.Write(bytearrayinput, 0, lu); 
      effectue += lu;
    }

    cryptostream.Close();
  }


  //-----------------------------------------------------------------------------------------------------------------
  //--                                           DecryptFluxIter                                                     --
  //-----------------------------------------------------------------------------------------------------------------
  ///<summary>Decrypt un flux encod en plusieurs passe</summary>
  ///<remark>A utiliser pour les flux de grande taille. Les flux sont NON ferms  la fin de l'exec</remark>
  private static void DecryptFluxIter (AKKPassword s, Stream sIN, Stream sOUT){
    DESCryptoServiceProvider DES = new DESCryptoServiceProvider(); 

    //DES utilise une cl (pwd) & un I(nitialization) V(ector) : mme contraire = 8 char
    DES.Key = s.PwdBytes();
    DES.IV = s.PwdBytes();

    ICryptoTransform dk = DES.CreateDecryptor();
    CryptoStream cryptostream = new CryptoStream(sIN,
						 dk, 
						 CryptoStreamMode.Read); 
    
    //** la diffrence est ici : on fait des itrations ... **
    
    //le buffer sera de taille multiple de la taille de la cl (8 octets) -> 16384
    byte[] bytearrayinput = new byte[16384];
    int lu; //car le fichier encrypt a une taille modulo 8 qui peut tre suprieure  celle du fichier d'origine !!!
    do {      
      lu = cryptostream.Read(bytearrayinput, 0, bytearrayinput.Length);
      sOUT.Write(bytearrayinput, 0, lu); 
    } while (lu != 0);

    cryptostream.Close();
  }


}




///<summary>Mot de passe au bon format</summary>
///<remark>Utilisant DES, le mot de passe doit faire 8 caractres (pas +, pas -)</remark>
public class AKKPassword{
  protected String pwd;
  public String Pwd {
    get{
      return pwd;
    }
  }

  protected AKKPassword(){}

  public AKKPassword (String p){
    if (p.Length != 8)
      throw new InvalidDESPassword();
    pwd = p;
  }

  public byte[] PwdBytes(){
    return ASCIIEncoding.ASCII.GetBytes(pwd);
  }
}


///<summary>Mot de passe TRANSFORME au bon format</summary>
///<remark>Utilisant DES, le mot de passe doit faire 8 caractres (pas +, pas -). Cette sous classe s'arrange pour gnrer toujours de la mme manire un mdp de 8 caractres avec un qui fait - OU +.</remark>
public class AKKPasswordHarmonized : AKKPassword{

  public AKKPasswordHarmonized (String p){
    if (p.Length == 0)
      throw new InvalidDESPassword();

    if (p.Length == 8)
      pwd = p;
    else{ //on bidouille
      if (p.Length > 8)
	pwd = p.Substring (1, 8); //ex : "azertyuiop" deviens "zertyuio"
      else{
	//plus petit : on repete le motif
	pwd = p;
	int i = 0;
	do {
	  pwd = pwd + p.Substring(i, 1);
	  i = (i + 1) % p.Length;
	} while (pwd.Length != 8);
      }
	
    }
  }
}



///<summary>Levee si le mot de passe donn n'est pas un mot de passe DES valide</summary>
public class InvalidDESPassword : ApplicationException {
  public InvalidDESPassword() : base ("Le mot de passe fournis n'est pas un mot de passe DES valide") {}
}
///<summary>Levee si le mot de passe donn n'est pas celui qui a servi  cre le flux crypt</summary>
public class BadPassword : ApplicationException {
  public BadPassword() : base ("Le mot de passe fournis n'est pas celui ayant servi  encoder le flux") {}
}
